import asyncio
from aiohttp import web, streams
from gera2ld.socks.client import create_client
from .util import forward_data

class ConnectReader:
    def __init__(self, queue):
        self.queue = queue
        self._exc = None

    def feed_eof(self):
        self.queue.feed_eof()

    def feed_data(self, data):
        if self._exc:
            return True, data
        try:
            self.queue.feed_data(data, len(data))
            return False, b''
        except Exception as exc:
            self._exc = exc
            self.queue.set_exception(exc)
            return True, b''

async def handle(handler, request):
    host, _, port = request.raw_path.partition(':')
    port = int(port)
    if handler.socks_proxy:
        client = create_client(handler.socks_proxy, remote_dns=True)
        await client.handle_connect((host, port))
        reader, writer = client.reader, client.writer
    else:
        reader, writer = await asyncio.open_connection(host, port)
    response = web.StreamResponse(
        status=200,
        reason='Connection established',
    )
    # disable chunked encoding
    response._length_check = False
    await response.prepare(request)
    response._reader = streams.DataQueue(loop=request._loop)
    request.protocol.set_parser(ConnectReader(response._reader))
    request.protocol.keep_alive(False)
    async def forward_client_data():
        async for chunk in response._reader:
            writer.write(chunk)
            await writer.drain()
        reader.feed_eof()
    try:
        await asyncio.wait([
            forward_client_data(),
            forward_data(reader, response),
        ])
    except asyncio.CancelledError:
        # Connection lost causes cancellation of current task,
        # we catch CancelledError so that access will still be logged.
        pass
    return response



from aiohttp import web
from .connect import handle as handle_connect
from .proxy import handle as handle_proxy

class Handler:
    def __init__(self, socks_proxy=None):
        self.socks_proxy = socks_proxy

    async def __call__(self, request):
        if request.method == 'CONNECT':
            return await handle_connect(self, request)
        if '://' in request.raw_path:
            return await handle_proxy(self, request)
        if request.method not in ('HEAD', 'GET'):
            raise web.HTTPMethodNotAllowed
        raise web.HTTPNotFound



__author__ = """Gerald"""

try:
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata

try:
    __version__ = metadata.version('json-server.py')
except:
    __version__ = 'DEV'



import asyncio
import aiohttp
import socket
from typing import Any, Dict, List, Optional, Tuple, Type
from aiohttp import web, TCPConnector
from aiohttp.helpers import get_running_loop
from aiohttp.client_proto import ResponseHandler
from gera2ld.socks.client import create_client
from .util import forward_data

class SOCKSConnector(TCPConnector):
    def __init__(self, socks_proxy, *k, **kw):
        super().__init__(*k, **kw)
        self.socks_proxy = socks_proxy

    async def _create_connection(self, req: 'ClientRequest',
                                 traces: List['Trace'],
                                 timeout: 'ClientTimeout') -> ResponseHandler:
        client = create_client(self.socks_proxy, remote_dns=True)
        await client.handle_connect((req.url.host, req.url.port))
        rawsock = client.writer._transport.get_extra_info('socket', default=None)
        self.client = client
        _transp, proto = await self._wrap_create_connection(
            self._factory, timeout=timeout,
            sock=rawsock,
            req=req)
        return proto

    def _release(self, *k, **kw) -> None:
        self.client.writer.close()
        return super()._release(*k, **kw)

async def handle(handler, request):
    connector = SOCKSConnector(handler.socks_proxy, force_close=True)
    async with aiohttp.ClientSession(connector=connector, auto_decompress=False) as session:
        async with session.request(
            request.method,
            request.raw_path,
            headers=request.headers,
            data=await request.content.read(),
            allow_redirects=False,
        ) as client_response:
            headers = client_response.headers.copy()
            headers.popall('transfer-encoding', None) # e.g. chunked
            for key in headers.keys():
                if key.lower().startswith('proxy-'):
                    headers.popall(key, None)
            response = web.StreamResponse(
                status=client_response.status,
                reason=client_response.reason,
                headers=headers,
            )
            await response.prepare(request)
            await forward_data(client_response.content, response)
            return response



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gera2ld', 'gera2ld.aioproxy', 'gera2ld.aioproxy.handlers']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.6.2,<4.0.0',
 'async_dns>=1.0.10,<2.0.0',
 'click>=7.0,<8.0',
 'gera2ld.socks>=0.2.0,<0.3.0',
 'importlib_metadata>=1.7.0,<2.0.0']

entry_points = \
{'console_scripts': ['aioproxy = gera2ld.aioproxy.cli:main']}

setup_kwargs = {
    'name': 'gera2ld.aioproxy',
    'version': '0.1.0',
    'description': '',
    'long_description': '# gera2ld.aioproxy\n\n[![PyPI](https://img.shields.io/pypi/v/gera2ld.aioproxy.svg)](https://pypi.org/project/gera2ld.aioproxy/)\n\nA simple HTTP proxy server implemented with pure Python on top of [aiohttp](https://aiohttp.readthedocs.io/).\n\n## Installation\n\n```sh\n$ pip3 install gera2ld.aioproxy\n```\n\n## Usage\n\n```sh\n$ aioproxy [-b :1086] [-x socks5://127.0.0.1:1080]\n```\n',
    'author': 'Gerald',
    'author_email': 'gera2ld@163.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gera2ld/aioproxy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import os
import sys
import platform
import logging
import click
from gera2ld.pyserve import run_forever, start_server_aiohttp
from aiohttp.log import server_logger
from . import __version__
from .handlers import Handler

@click.command()
@click.option('-b', '--bind', default=':1086', help='the server address to bind')
@click.option('-x', '--proxy', default='socks5://127.0.0.1:1080', help='downstream SOCKS proxy')
def main(bind, proxy):
    logging.basicConfig(level=logging.INFO)
    server_logger.info(
        'Proxy Server v%s/%s %s - by Gerald',
        __version__, platform.python_implementation(), platform.python_version())
    run_forever(start_server_aiohttp(Handler(proxy), bind))
    return 0

if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



import asyncio

BUF_SIZE = 4096

async def forward_data(reader, writer):
    try:
        while True:
            try:
                chunk = await reader.read(BUF_SIZE)
            except ConnectionResetError:
                break
            if not chunk:
                break
            # https://github.com/aio-libs/aiohttp/issues/3122#issuecomment-567960647
            await asyncio.shield(writer.write(chunk))
    finally:
        await writer.write_eof()
